require 'main/container'
require 'main/types'

Figlet = :DiskmanFiglet

### Misc constants
UNSET = 'unset'
RAID  = 'raid'
PLAIN = 'plain'
PROC_CPUINFO = '/proc/cpuinfo'
BIN_EVMS = '/usr/sbin/evms'

class DiskmanFiglet < Core::Figlet
    attr_reader :evms, :volumes_root
    def initialize
	super("disk_manager")
	@filenames = [] #["diskman.dump"]
	@base_path = "/var/cache/rconf/"
	@evms = Evms.new
	
	@volumes_root = '/drives/'
	
	File.foreach('/etc/rconf.conf') {|line|
	    @volumes_root = line.sub(/.*\:/, '').strip! if line =~/volume_root.*/
	} if File.exists? '/etc/rconf.conf'
	
	## constructing the hierarchy
	@hierarchy = Core::Notebook.new("disk_tree") 
	@hierarchy.residence = self
	

	load
    end

    def add_entities (kind_name, prnt)
	(@evms.send kind_name).each do |item|
	    data = @evms.get_data(item, kind_name)
	    #puts "And the data is: #{data} for item #{item}\n"
	    item.strip!
	    case kind_name
	    when :info_containers then o = EVMSContainer.new("", prnt, data)
	    when :info_segments then o = EVMSSegment.new("", prnt, data)
	    when :info_volumes then o = EVMSVolume.new("", prnt, data)
	    when :info_disks then o = EVMSDisk.new("", prnt, data)
	    when :info_regions then o = EVMSRegion.new("", prnt, data)
	    else raise "error at DiskmanFiglet#add_entities: " + kind_name.to_s + "\n"
	    end #case
	    o.label = item
	    @entities[item] = o

	    unless o.class == Diskman::EVMSDisk
		pars = @evms.info('children, ' + item, ' Name:')
		if pars.length > 0
		    list = EVMSList.new('parents_list', o) 
		    pars.each do |par|
			par.strip!
			Core::Hyperlink.new('', list) {@entities[par]}
		    end
		end
	    end

	    childrn = @evms.info('parent, ' + item, ' Name:')
	    if childrn.length > 0
		list = EVMSList.new('children_list', o)
		childrn.each do |child|
		    child.strip!
		    #puts "Child of " + item + " is " + child
		    Core::Hyperlink.new('', list){@entities[child]}
		end
	    end
	end
    end

    def save
	
	@evms.exec_evms "Commit"
	#File.open(@base_path + @filenames[0], 'w'){|file| Marshal.dump(@hierarchy, file)}

    end
    
    def load

	@hierarchy.children.clear

	## creating a hash where all our entities will float
	@entities = Hash.new

	##adding available containers
	containers = ObjectList.new("disk_containers", @hierarchy)
	
	add_entities (:info_containers, containers)

	##adding available volumes
	volumes = ObjectList.new("disk_volumes", @hierarchy)
	
	add_entities (:info_volumes, volumes)
	Core::Action.new("create", volumes) do |y|
	    page = Core::Notebook.new("creation", volumes)
	    objects = @entities.values.find_all{|obj|
		#puts "obj.class: #{obj.class}"
		## check if obj is among suitable volume substrates
		[Diskman::EVMSDisk, Diskman::EVMSSegment, Diskman::EVMSRegion, Diskman::EVMSContainer].index obj.class}
	    choices = Array.new
	    objects.each {|obj| choices.push "#{obj.label}: #{obj.class.to_s.sub!(/.*\:\:EVMS/, '')}"}
	    if choices.length > 0
		base_object = Core::Variant.new("base_object", choices, page)
		volume_name = Core::Leaf.new("volume_name", page)
		Core::Action.new("volume_create", page){|x|
		    # @evms.exec_evms("c: v, #{base_object.val.sub(/\:.*/, '')}, name = #{volume_name.val}")
		    #puts "c: v, #{x[base_object].sub(/\:.*/, '')}, name = #{x[volume_name]}"
		    ## creating the volume
		    @evms.exec_evms("c: v, #{x[base_object].sub(/\:.*/, '')}, name = #{x[volume_name]}")
		    system("/sbin/mkfs.xfs /dev/evms/#{x[volume_name]}") && system("mkdir #{volumes_root}#{x[volume_name]}") && system("mount -t xfs /dev/evms/#{x[volume_name]} #{@volumes_root}#{x[volume_name]}")
		    
		    load
		    # @hierarchy
		    page.ref "/top/disk_tree/disk_volumes"}
		page
	    else
		Core::Message.new("disk_no_volumes", nil, "No base objects available")
	    end
	end
	
	##adding available segments
	segments = ObjectList.new("disk_segments", @hierarchy)
	
	add_entities (:info_segments, segments)
	Core::Action.new("create", segments) do |y|
	    page = Core::Notebook.new("creation", segments)
	    objects = @entities.values.find_all{|obj|
		## check if obj is among suitable segment substrates
		obj.class == Diskman::EVMSSegment && /Freespace/.match(obj.segment_type)
	    }
	    choices = Array.new
	    objects.each {|obj| choices.push "#{obj.label}: Freespace Segment"}
	    if choices.length > 0
		base_object = Core::Variant.new("base_object", choices, page)
		segment_size = Core::Number.new("segment_size", page, 0)
		Core::Action.new("segment_create", page){|x|
		    @evms.exec_evms("c: s, #{x[base_object].sub(/\:.*/, '')}, size=#{x[segment_size].to_s}MB, primary=false, bootable=false")
		    load
		    # @hierarchy
		    page.ref "/top/disk_tree/disk_segments"
		}
		page
	    else
		Core::Message.new("disk_no_freespace", nil, "No freespace segments available")
	    end
	end

	##adding available disks
	disks = ObjectList.new("disk_disks", @hierarchy)
	
	add_entities (:info_disks, disks)
	
	##adding available regions
	regions = ObjectList.new("disk_regions", @hierarchy)
	
	add_entities (:info_regions, regions)
	
	## refresh button
	Core::Action.new("redraw", @hierarchy) do |x|
	    @evms.probe #update the evms state
	    load
	    @hierarchy
	end    
	
	##adding drive hot-plug button
	

    end #load
end

class EVMSList < Core::MutableList

    ## accept only hyperlinks to EVMS nodes
    def initialize (name, parent)
	super(name, parent) {|x| Core::Hyperlink===x}
    end

    #	def actionAdd
    #	end

end

### we use it to list our manageable objects
class ObjectList < Core::MutableList
    def initialize (name, parent)
	super (name, parent) {|x| Diskman::EVMSNode===x}
    end
end

### the set of classes for storing data about EVMS elements
class EVMSNode < Core::Notebook
    
    def initialize (name, parent, data)
	super(name, parent)
	@data = data
    end

    def actionDelete
	@residence.evms.exec_evms "dr: #{self.label}"
    end

    
end #EVMSNode

class EVMSContainer < EVMSNode
end

class EVMSDisk < EVMSNode
    undef_method :actionDelete
    def initialize(name, parent, data)
	super(name, parent, data)
	result = @data.find{|str| /.*Size\: *(.*)/.match(str)}
	Core::Text.new('', self, result)# if matched

	result = @data.find{|str| /.*Geometry\: *(.*)/.match(str)}
	Core::Text.new('', self, result)
    end
end

class EVMSSegment < EVMSNode
    attr_reader :segment_type
    def initialize(name, parent, data)
	super(name, parent, data)
	result = @data.find{|str| /.*Size\: *(.*)/.match(str)}
	Core::Text.new('', self, result)# if matched

	result = @data.find{|str| /.*LBA\: *(.*)/.match(str)}
	Core::Text.new('', self, result)

	result = @data.find{|str| /.*Type\: *(.*)/.match(str)}
	Core::Text.new('', self, result)
	@segment_type = result
	#puts @segment_type
    end
end

class EVMSVolume < EVMSNode
    #undef_method :edit
    def initialize(name, parent, data)
	super(name, parent, data)
	result = @data.find{|str| /.*Size\: *(.*)/.match(str)}
	Core::Text.new('', self, result)# if matched

	result = @data.find{|str| /.*Mount Point\: *(.*)/.match(str)}
	Core::Text.new('', self, result)

	result = @data.find{|str| /.*Filesystem\: *(.*)/.match(str)}
	if result
	    Core::Text.new('', self, result)
	    Core::Text.new('', self, @data.find{|str| /.*Min Filesystem Size\: *(.*)/.match(str)})
	    Core::Text.new('', self, @data.find{|str| /.*Max Filesystem Size\: *(.*)/.match(str)})
	else
	    Core::Text.new('', self, "Filesystem: unknown type or unfromatted")
	end #if
    end

    ## we handle special case of non-deletable objects (e.g. /root, swap) here
    def label= value
	super
	instance_eval "class <<self; private :actionDelete; end" if ["/dev/evms/swap", "/dev/evms/root", "/dev/evms/Volume1"].index value
    end

    def actionDelete
	## unmount and destroy the mount point
	#system "unalias rm"
	error = false
	system_guarded("umount #{@residence.volumes_root}#{self.label.sub(/\/dev\/evms\//, '')}")
	system_guarded("rmdir #{@residence.volumes_root}#{self.label.sub(/\/dev\/evms\//, '')}")
	super
    end
end

class EVMSRegion < EVMSNode
    def initialize(name, parent, data)
	super(name, parent, data)
	result = @data.find{|str| /.*Size\: *(.*)/.match(str)}
	Core::Text.new('', self, result)# if matched
	
	result = @data.find{|str| /.*Type\:(.*)/.match(str)}
	Core::Text.new('', self, result)
    end
end

## Aux. classes for EVMS CLI interaction
class Evms
    def wait_resync
	while true do
	    results = []    
	    File.open("/proc/evms/mdstat","r") do |f|
		lines = f.readlines
		raid = nil
		results = lines.collect do |line|
		    case line
		    when /^\s*(md\d) :/
			raid = $1
			nil
		    when /(finish|speed)/
			if raid
			    resync = nil
			    recovery = nil
			    finish = nil
			    speed = nil
			    if line =~ / resync =[ ]*(\d+\.\d)%/;    resync = $1;   end
			    if line =~ / recovery =[ ]*(\d+\.\d)%/;  recovery = $1; end
			    if line =~ / finish=(\d+\.\d)min/;       finish = $1; 	end
			    if line =~ /speed=(\d+K\/sec)/;		 speed = $1;	end
			    araid = raid
			    raid = nil
			    [araid,resync.to_i,recovery.to_i,finish,speed]
			else
			    nil
			end
		    else
			nil
		    end
		end
	    end
	    results.compact!
	    if !results.empty?
		results.each do |raid,resync,recovery,finish,speed|
		    gauge.gauge_update(_("RAID #{raid}: resyncing and recovering [#{speed}] ETA = #{finish} min"), resync, recovery)
		end
	    else
		break
	    end
	    sleep(2)
	end
    end

    def drive_names(); return @drives.keys.to_a; end
    def drive_size(drive); return @drives[drive]['size']; end

    def initialize()
	get_info();
    end
    
    def get_info()
	@drives = get_drives_info()
    end

    def probe
	exec_evms "Probe"
    end

    def get_data(name, kind_name)
	tran = {:info_volumes => 'v',
	    :info_regions => 'r',
	    :info_disks => 'd',
	    :info_segments => 's',
	    :info_containers => 'c'}
	
	data = UNSET
	cmd = "q: " + tran[kind_name]

	data=exec_evms(cmd).join
	data.gsub!(/\n/, '#')
	data.sub!(/^.*Name\: *#{name}\#/, '') if data =~/Name\: *#{name}\#/
	data.gsub!(/\#\#.*EVMS\:/, '') if data =~/\#\#/
	data.gsub!(/\#/, "\n")
	#(! rw.closed?) && rw.close
	return data
    end

    def exec_evms(cmd)
	result = Array.new;
	data = UNSET;
	rw = IO.popen("#{BIN_EVMS}", "r+");
	while ((! rw.closed?) && ((res = IO.select([rw],nil,nil,120)) != nil))
	    begin
		(! rw.eof) && (data = rw.readline) && result.push(data);
		if ((data =~ "EVMS:") && (cmd != ''))
		    rw.write("#{cmd}\n");
		    cmd = '';
		elsif ((data =~ 'Please enter the number corresponding to your choice:') or (data =~ 'The following responses are available:'))
		    rw.write("1\n");
		elsif (data =~ "EVMS:")
		    rw.write("commit\n");
		    rw.write("quit\n");
		    rw.close;
		end
	    rescue
		rw.close;
	    end
	end;
	(! rw.closed?) && rw.close;
	return result;
    end
    
    def info(query_type, grep_str)
	result = Array.new;
	exec_evms("Query:#{query_type}").each{|str| str.chop!;
	    if (str =~ /#{grep_str}/)
		str.gsub!(/^.*#{grep_str}/, '');
		result.push(str);
	    end
	}
	return result;
    end
    
    def info_regions();     return info('R', 'Region Name: ');    end
    def info_volumes();     return info('V', 'Volume Name: ');    end
    def info_segments();    return info('S', 'Segment Name: ');   end
    def info_containers();  return info('C', 'Container Name: '); end
    def info_disks(); return info('D', 'Disk Name: '); end
    def info_disks_unclaimed
	info('D, U', 'Disk Name: ')
    end
    def assign_drv(device); exec_evms("Assign:DosSegMgr={TypeByName=Linux},#{device.gsub(/\/dev\//, '')}"); end

    def create_segment(drive, size, bootf='false')
	exec_evms("Create:S,#{drive.gsub(/\/dev\//, '')}_freespace1,size=#{size}MB,primary=true,bootable=#{bootf}");
    end
    def create_region(drives, type);  exec_evms("Create:R,MD#{type}RegMgr={},#{drives}"); end
    def create_volume(segment, name); exec_evms("Create:V,#{segment},Name=\"#{name}\"");  end

    def remove_volume(volume)
	volume.sub!('/dev/evms/', '')
	exec_evms("Revert:/dev/evms/#{volume}")
	exec_evms("Delete:#{volume}")
	exec_evms("Delete:/dev/evms/#{volume}")
    end

    def clean_drives()
	segments_list = Array.new;
	info_segments().each  {|segment|  (segment !~ 'freespace') && segments_list.push(segment)}
	info_volumes().each   {|volume|    remove_volume(volume)}

	# First remove LVM regions, because lvm may to be on md/mdX regions
	info_regions().each   {|region|    (region =~ /lvm/) && exec_evms("Delete:#{region}");}
	info_containers().each{|container| exec_evms("Delete:#{container}");}

	# Delete other regions
	info_regions().each   {|region|    exec_evms("Delete:#{region}");}
	segments_list.each    {|segment|   exec_evms("Delete:#{segment}");}

    end

    def get_drives_info()
	drives = Hash.new(); drive = '';
	exec_evms("Query:D").each {|str| str.chop!;
	    if str =~ /Disk\ Name: (.*)/;
		drive = $1;
	    end
	    if str =~ /Disk\ Size: (.*) ?[MG][Bb]/
		drives[drive] = {'size' => $1}
	    end
	}
	return drives;
    end

    def mkdir(dirs)
	dirs.split(',').each{|dir|
	    begin
		File.makedirs(dir, false);
	    rescue
	    end
	}
    end

    def create_dirs(top_dir, dir_list)
	dir_list.each_pair{|dir, perm|
	    real_dir = top_dir+dir;
	    mkdir(real_dir);
	    File.chmod(perm, real_dir);
	}
    end

    def find_file(dir, filename)
	return Dir["#{dir}/#{filename}*"]
    end

end    
